from typing import Optional
from guardpost.authentication import Identity


class OperationContext:

    __slots__ = ('identity',)

    def __init__(self):
        self.identity = None  # type: Optional[Identity]




import ntpath
import pathlib
from typing import Tuple
from mimetypes import MimeTypes

mime = MimeTypes()


DEFAULT_MIME = 'application/octet-stream'

MIME_BY_EXTENSION = {
    '.ogg': 'audio/ogg',
    '.jpg': 'image/jpeg',
    '.jpeg': 'image/jpeg',
    '.png': 'image/png',
    '.woff2': 'font/woff2'
}


def get_file_extension_from_name(name: str) -> str:
    if not name:
        return ''

    extension = pathlib.Path(name).suffix
    return extension.lower()


def get_file_name_from_path(full_path: str) -> str:
    head, tail = ntpath.split(full_path)
    return tail or ntpath.basename(head)


def get_mime_type(file_name: str) -> bytes:
    extension = get_file_extension_from_name(file_name)
    mime_type = mime.guess_type(file_name)[0] or DEFAULT_MIME

    if mime_type == DEFAULT_MIME and extension in MIME_BY_EXTENSION:
        mime_type = MIME_BY_EXTENSION.get(extension)

    return mime_type.encode()


def get_best_mime_type(file_name: str) -> Tuple[str, str]:
    extension = get_file_extension_from_name(file_name)
    mime_type = mime.guess_type(file_name)[0] or DEFAULT_MIME

    if mime_type == DEFAULT_MIME and extension in MIME_BY_EXTENSION:
        mime_type = MIME_BY_EXTENSION.get(extension)

    return extension, mime_type



from .messages cimport Request, Response
from .contents cimport TextContent, HtmlContent
from .exceptions cimport HttpException, NotFound


import os
import html
import traceback


async def handle_not_found(app, Request request, HttpException http_exception):
    return Response(404, content=TextContent('Resource not found'))


async def handle_bad_request(app, Request request, HttpException http_exception):
    return Response(400, content=TextContent(f'Bad Request: {str(http_exception)}'))


cdef class BaseApplication:

    def __init__(self, bint show_error_details, object router, object services):
        self.router = router
        self.services = services
        self.connections = []
        self.exceptions_handlers = self.init_exceptions_handlers()
        self.show_error_details = show_error_details

    def init_exceptions_handlers(self):
        return {
            404: handle_not_found,
            400: handle_bad_request
        }

    async def handle(self, Request request):
        cdef object route
        cdef Response response

        route = self.router.get_match(request.method, request._path)

        if route:
            request.route_values = route.values

            try:
                response = await route.handler(request)
            except HttpException as http_exception:
                response = await self.handle_http_exception(request, http_exception)
            except Exception as exc:
                response = await self.handle_exception(request, exc)
        else:
            response = await self.exceptions_handlers.get(404)(self, request, None)
            if not response:
                response = Response(404)
        # if the request handler didn't return an object,
        # and since the request was handled successfully, return success status code No Content
        # for example, a user might return "None" from an handler
        # this might be ambiguous, if a programmer thinks to return None for "Not found"
        return response or Response(204)

    cdef object get_http_exception_handler(self, HttpException http_exception):
        return self.exceptions_handlers.get(http_exception.status)

    cdef object get_exception_handler(self, Exception exception):
        return self.exceptions_handlers.get(type(exception))

    async def handle_internal_server_error(self, Request request, Exception exc):
        if self.show_error_details:
            tb = traceback.format_exception(exc.__class__,
                                            exc,
                                            exc.__traceback__)
            info = ''
            for item in tb:
                info += f'<li><pre>{html.escape(item)}</pre></li>'

            content = HtmlContent(self.resources.error_page_html
                                  .format_map({'process_id': os.getpid(),
                                               'info': info,
                                               'exctype': exc.__class__.__name__,
                                               'excmessage': str(exc),
                                               'method': request.method,
                                               'path': request.url.value.decode()}))

            return Response(500, content=content)
        return Response(500, content=TextContent('Internal server error.'))

    async def _apply_exception_handler(self, Request request, Exception exc, object exception_handler):
        try:
            return await exception_handler(self, request, exc)
        except Exception as server_ex:
            return await self.handle_exception(request, server_ex)

    async def handle_http_exception(self, Request request, HttpException http_exception):
        exception_handler = self.get_http_exception_handler(http_exception)
        if exception_handler:
            return await self._apply_exception_handler(request, http_exception, exception_handler)

        return await self.handle_exception(request, http_exception)

    async def handle_exception(self, request, exc):
        exception_handler = self.get_exception_handler(exc)
        if exception_handler:
            return await self._apply_exception_handler(request, exc, exception_handler)

        return await self.handle_internal_server_error(request, exc)



import logging
from typing import Optional
from ipaddress import ip_address
from datetime import datetime, timedelta
from blacksheep import Request, Cookie, URL


client_logger = logging.getLogger('blacksheep.client')


class InvalidCookie(Exception):

    def __init__(self, message):
        super().__init__(message)


class InvalidCookieDomain(InvalidCookie):

    def __init__(self):
        super().__init__('Invalid domain attribute')


class MissingHostInURL(ValueError):

    def __init__(self):
        super().__init__('An URL with host is required.')


def not_ip_address(value: str):
    try:
        ip_address(value)
    except ValueError:
        return True
    return False


class StoredCookie:

    __slots__ = ('cookie',
                 'persistent',
                 'creation_time',
                 'expiry_time')

    def __init__(self, cookie: Cookie):
        # https://tools.ietf.org/html/rfc6265#section-5.3
        self.cookie = cookie
        self.creation_time = datetime.utcnow()

        expiry = None
        if cookie.max_age:
            # https://tools.ietf.org/html/rfc6265#section-5.2.2
            try:
                max_age = int(cookie.max_age)
            except ValueError:
                pass
            else:
                if max_age <= 0:
                    expiry = datetime.min
                else:
                    expiry = self.creation_time + timedelta(seconds=max_age)
        elif cookie.expires:
            expiry = cookie.expiration

        self.expiry_time = expiry
        self.persistent = True if expiry else False

    @property
    def name(self):
        return self.cookie.name

    def is_expired(self) -> bool:
        expiration = self.expiry_time
        if expiration and expiration < datetime.utcnow():
            return True

        # NB: it's a 'session cookie'; in other words
        # it expires when the session is closed
        return False


class CookieJar:

    def __init__(self):
        self._domain_cookies = {}     # cookies with specific domain
        self._host_only_cookies = {}  # cookies without specific domain

    @staticmethod
    def _get_url_host(request_url: URL):
        if not request_url.host:
            raise MissingHostInURL()

        return request_url.host.lower()

    @staticmethod
    def _get_url_path(request_url: URL):
        if request_url.path:
            return request_url.path.lower()
        return b'/'

    def get_domain(self, request_url: URL, cookie: Cookie) -> bytes:
        # https://tools.ietf.org/html/rfc6265#section-4.1.2.3
        request_domain = self._get_url_host(request_url)

        if not cookie.domain:
            return request_domain

        cookie_domain = cookie.domain.lstrip(b'.').lower()

        if cookie_domain.endswith(b'.'):
            # ignore the domain attribute;
            return request_domain

        if not request_domain.endswith(cookie_domain):
            client_logger.warning(f'A response for {request_url.value} tried to set '
                                  f'a cookie with domain {cookie_domain}; this could '
                                  f'be a malicious action.')
            raise InvalidCookieDomain()

        return cookie_domain

    @staticmethod
    def get_path(request_url: URL, cookie: Cookie):
        if cookie.path:
            return cookie.path.lower()

        return CookieJar.get_cookie_default_path(request_url)

    @staticmethod
    def get_cookie_default_path(request_url: URL) -> bytes:
        # https://tools.ietf.org/html/rfc6265#section-5.1.4
        uri_path = request_url.path

        if not uri_path or not uri_path.startswith(b'/'):
            return b'/'

        if uri_path == b'/':
            return uri_path

        return uri_path[0:uri_path.rfind(b'/')]

    @staticmethod
    def domain_match(domain: bytes, value: bytes):
        lower_domain = domain.lower()
        lower_value = value.lower()

        if lower_domain == lower_value:
            return True

        return lower_value.startswith(lower_domain) \
               and lower_value[len(lower_domain)] == 46 \
               and not_ip_address(lower_value.decode())

    @staticmethod
    def path_match(request_path: bytes, cookie_path: bytes):
        # https://tools.ietf.org/html/rfc6265#section-5.1.4
        lower_request_path = request_path.lower()
        lower_cookie_path = cookie_path.lower()

        if lower_request_path == lower_cookie_path:
            return True

        if lower_request_path.startswith(lower_cookie_path) \
                and lower_cookie_path[-1] == 47:
            return True

        if lower_request_path.startswith(lower_cookie_path) \
                and lower_request_path[len(lower_cookie_path)] == 47:
            return True

        return False

    def get_cookies_for_url(self, url: URL):
        return self.get_cookies(url.schema,
                                self._get_url_host(url),
                                self._get_url_path(url))

    def _get_cookies_by_path(self, schema: bytes, path: bytes, cookies_by_path: dict):
        for cookie_path, cookies in cookies_by_path.items():
            if CookieJar.path_match(path, cookie_path):
                for cookie in self._check_cookies(schema, cookies):
                    yield cookie.clone()

    @staticmethod
    def _check_cookies(schema: bytes, cookies: dict):
        for cookie_name, stored_cookie in cookies.copy().items():

            if stored_cookie.is_expired():
                try:
                    del cookies[cookie_name]
                except KeyError:
                    pass
                continue

            cookie = stored_cookie.cookie
            if cookie.secure and schema != b'https':
                # skip cookie for this request
                continue

            yield cookie

    def get_cookies(self, schema: bytes, domain: bytes, path: bytes):
        for cookies_domain, cookies_by_path in self._host_only_cookies.items():
            if cookies_domain == domain:
                yield from self._get_cookies_by_path(schema, path, cookies_by_path)

        for cookies_domain, cookies_by_path in self._domain_cookies.items():
            if CookieJar.domain_match(cookies_domain, domain):
                yield from self._get_cookies_by_path(schema, path, cookies_by_path)

    @staticmethod
    def _ensure_dict_container(container, key):
        try:
            return container[key]
        except KeyError:
            new_container = {}
            container[key] = new_container
            return new_container

    def _set_ensuring_container(self, root_container, domain, path, stored_cookie: StoredCookie):
        domain_container = self._ensure_dict_container(root_container, domain)
        path_container = self._ensure_dict_container(domain_container, path)
        domain_container[path] = path_container
        path_container[stored_cookie.name.lower()] = stored_cookie

    @staticmethod
    def _get(container: dict, domain: bytes, path: bytes, cookie_name: bytes) -> Optional[StoredCookie]:
        try:
            return container[domain][path][cookie_name]
        except KeyError:
            return None

    @staticmethod
    def _remove(container: dict, domain: bytes, path: bytes, cookie_name: bytes) -> bool:
        try:
            del container[domain][path][cookie_name]
        except KeyError:
            return False
        return True

    def get(self, domain: bytes, path: bytes, cookie_name: bytes) -> Optional[StoredCookie]:
        return self._get(self._host_only_cookies, domain, path, cookie_name) \
               or self._get(self._domain_cookies, domain, path, cookie_name)

    def remove(self, domain: bytes, path: bytes, cookie_name: bytes) -> bool:
        return self._remove(self._host_only_cookies, domain, path, cookie_name) \
               or self._remove(self._domain_cookies, domain, path, cookie_name)

    def add(self, request_url: URL, cookie: Cookie):
        domain = self.get_domain(request_url, cookie)
        path = self.get_path(request_url, cookie)
        cookie_name = cookie.name.lower()

        stored_cookie = StoredCookie(cookie)

        if cookie.domain:
            container = self._domain_cookies
        else:
            container = self._host_only_cookies

        # handle existing cookie
        # https://tools.ietf.org/html/rfc6265#page-23
        existing_cookie = self.get(domain, path, cookie_name)

        if existing_cookie:
            if existing_cookie.cookie.http_only and not cookie.http_only:
                # ignore
                return
            stored_cookie.creation_time = existing_cookie.creation_time

        if stored_cookie.is_expired():
            # remove existing cookie with the same name; if applicable;
            if existing_cookie:
                self.remove(domain, path, cookie_name)
            return

        self._set_ensuring_container(container, domain, path, stored_cookie)


async def cookies_middleware(request, next_handler):
    cookie_jar = request.context.cookies

    for cookie in cookie_jar.get_cookies_for_url(request.url):
        request.set_cookie(cookie)

    response = await next_handler(request)

    if b'set-cookie' in response.headers:
        for cookie in response.cookies.values():
            try:
                cookie_jar.add(request.url, cookie)
            except InvalidCookie as invalid_cookie_error:
                client_logger.debug(f'Rejected cookie for {request.url}; the cookie is invalid: '
                                    f'{str(invalid_cookie_error)}')
    return response




cdef class InvalidOperation(Exception):

    def __init__(self, str message, object inner_exception=None):
        super().__init__(message)
        self.inner_exception = inner_exception


cdef class BadRequestFormat(Exception):

    def __init__(self, str message, object inner_exception=None):
        super().__init__(message)
        self.inner_exception = inner_exception


cdef class HttpException(Exception):

    def __init__(self, int status, str message = None):
        super().__init__(message)
        self.status = status


cdef class BadRequest(HttpException):

    def __init__(self, message=None):
        super().__init__(400, message)


cdef class NotFound(HttpException):

    def __init__(self):
        super().__init__(404)


cdef class InvalidArgument(Exception):

    def __init__(self, str message):
        super().__init__(message)


cdef class MessageAborted(Exception):
    def __init__(self):
        super().__init__('The message was aborted before the client sent its whole content.')





def copy_special_attributes(source_method, wrapper):
    for name in {'auth', 'auth_policy', 'auth_schemes', 'allow_anonymous'}:
        if hasattr(source_method, name):
            setattr(wrapper, name, getattr(source_method, name))



from typing import Optional, Dict
from blacksheep import Response, TextContent
from guardpost.authorization import AuthorizationError
from guardpost.asynchronous.authentication import AuthenticationStrategy, AuthenticationHandler


__all__ = ('AuthenticationStrategy',
           'AuthenticationHandler',
           'AuthenticateChallenge',
           'get_authentication_middleware',
           'handle_authentication_challenge')


def get_authentication_middleware(strategy: AuthenticationStrategy):
    async def authentication_middleware(request, handler):

        await strategy.authenticate(request, getattr(handler, 'auth_schemes', None))

        return await handler(request)
    return authentication_middleware


class AuthenticateChallenge(AuthorizationError):
    header_name = b'WWW-Authenticate'

    def __init__(self,
                 scheme: str,
                 realm: Optional[str],
                 parameters: Optional[Dict[str, str]]):
        self.scheme = scheme
        self.realm = realm
        self.parameters = parameters

    def _get_header_value(self) -> bytes:
        if not self.realm and not self.parameters:
            return self.scheme.encode()

        parts = bytearray(self.scheme.encode())
        if self.realm:
            parts.extend(f' realm="{self.realm}"'.encode())

        if self.parameters:
            parts.extend(b', ')
            parts.extend(b', '.join([f'{key}="{value}"'.encode() for key, value in self.parameters.items()]))
        return bytes(parts)

    def get_header(self) -> tuple:
        return (self.header_name, self._get_header_value())


async def handle_authentication_challenge(app, request, exception: AuthenticateChallenge):
    return Response(401,
                    [exception.get_header()],
                    content=TextContent('Unauthorized'))



import os
import logging
from typing import Optional, List, Callable
from blacksheep.server.routing import Router
from blacksheep.server.logs import setup_sync_logging
from blacksheep.server.files.dynamic import serve_files
from blacksheep.server.resources import get_resource_file_content
from blacksheep.server.normalization import normalize_handler, normalize_middleware
from blacksheep.baseapp import BaseApplication
from blacksheep.messages import Request, Response
from blacksheep.contents import ASGIContent
from blacksheep.scribe import send_asgi_response
from blacksheep.middlewares import get_middlewares_chain
from blacksheep.server.authentication import (get_authentication_middleware,
                                              AuthenticateChallenge,
                                              handle_authentication_challenge)
from blacksheep.server.authorization import (get_authorization_middleware,
                                             AuthorizationWithoutAuthenticationError,
                                             handle_unauthorized)
from guardpost.authorization import UnauthorizedError, Policy
from guardpost.asynchronous.authentication import AuthenticationStrategy
from guardpost.asynchronous.authorization import AuthorizationStrategy
from rodi import Services


server_logger = logging.getLogger('blacksheep.server')


__all__ = ('Application',)


def get_default_headers_middleware(headers):
    async def default_headers_middleware(request, handler):
        response = await handler(request)
        for header in headers:
            response.headers.add(header)
        return response
    return default_headers_middleware


class Resources:

    def __init__(self, error_page_html):
        self.error_page_html = error_page_html


class ApplicationEvent:

    def __init__(self, context):
        self.__handlers = []
        self.context = context

    def __iadd__(self, handler):
        self.__handlers.append(handler)
        return self

    def __isub__(self, handler):
        self.__handlers.remove(handler)
        return self

    def __len__(self):
        return len(self.__handlers)

    def append(self, handler):
        self.__handlers.append(handler)

    async def fire(self, *args, **keywargs):
        for handler in self.__handlers:
            await handler(self.context, *args, **keywargs)


def get_show_error_details(show_error_details):
    if show_error_details:
        return True

    show_error_details = os.environ.get('BLACKSHEEP_SHOW_ERROR_DETAILS')

    if show_error_details and show_error_details not in ('0', 'false'):
        return True
    return False


class Application(BaseApplication):

    def __init__(self,
                 router: Optional[Router] = None,
                 middlewares: Optional[List[Callable]] = None,
                 resources: Optional[Resources] = None,
                 services: Optional[Services] = None,
                 debug: bool = False,
                 show_error_details: bool = False,
                 auto_reload: bool = True):
        if router is None:
            router = Router()
        if services is None:
            services = Services()
        super().__init__(get_show_error_details(debug or show_error_details), router, services)

        if middlewares is None:
            middlewares = []
        if resources is None:
            resources = Resources(get_resource_file_content('error.html'))
        self.debug = debug
        self.auto_reload = auto_reload
        self.running = False
        self.middlewares = middlewares
        self.processes = []
        self.access_logger = None
        self.logger = None
        self._default_headers = None
        self._use_sync_logging = False
        self._middlewares_configured = False
        self.resources = resources
        self._serve_files = None
        self._authentication_strategy = None  # type: Optional[AuthenticationStrategy]
        self._authorization_strategy = None  # type: Optional[AuthorizationStrategy]
        self.on_start = ApplicationEvent(self)
        self.on_stop = ApplicationEvent(self)
        self.started = False

    def use_authentication(self, strategy: Optional[AuthenticationStrategy] = None) -> AuthenticationStrategy:
        if self.running:
            raise RuntimeError('The application is already running, configure authentication '
                               'before starting the application')
        if not strategy:
            strategy = AuthenticationStrategy()

        self._authentication_strategy = strategy
        return strategy

    def use_authorization(self, strategy: Optional[AuthorizationStrategy] = None) -> AuthorizationStrategy:
        if self.running:
            raise RuntimeError('The application is already running, configure authorization '
                               'before starting the application')

        if not strategy:
            strategy = AuthorizationStrategy()

        if strategy.default_policy is None:
            # by default, a default policy is configured with no requirements,
            # meaning that request handlers allow anonymous users, unless specified otherwise
            # this can be modified, by adding a requirement to the default policy
            strategy.default_policy = Policy('default')

        self._authorization_strategy = strategy
        self.exceptions_handlers[AuthenticateChallenge] = handle_authentication_challenge
        self.exceptions_handlers[UnauthorizedError] = handle_unauthorized
        return strategy

    def route(self, pattern, methods=None):
        if methods is None:
            methods = ['GET']

        def decorator(f):
            for method in methods:
                self.router.add(method, pattern, f)
            return f
        return decorator

    def set_default_headers(self, headers):
        self._default_headers = headers

    def use_sync_logging(self):
        self._use_sync_logging = True

    def serve_files(self, folder_name: str, extensions=None, discovery=False, cache_max_age=10800):
        self._serve_files = folder_name, extensions, discovery, cache_max_age

    def _configure_sync_logging(self):
        logging_middleware, access_logger, app_logger = setup_sync_logging()
        self.logger = app_logger
        self.access_logger = access_logger,
        self.middlewares.insert(0, logging_middleware)

    def _apply_middlewares_in_routes(self):
        configured_handlers = set()

        for route in self.router:
            if route.handler in configured_handlers:
                continue

            route.handler = get_middlewares_chain(self.middlewares, route.handler)

            configured_handlers.add(route.handler)
        configured_handlers.clear()

    def _normalize_middlewares(self):
        self.middlewares = [normalize_middleware(middleware, self.services) for middleware in self.middlewares]

    def normalize_handlers(self):
        configured_handlers = set()

        for route in self.router:
            if route.handler in configured_handlers:
                continue

            route.handler = normalize_handler(route, self.services)

            configured_handlers.add(route.handler)
        configured_handlers.clear()

    def configure_middlewares(self):
        if self._middlewares_configured:
            return
        self._middlewares_configured = True

        if self._authorization_strategy:
            if not self._authentication_strategy:
                raise AuthorizationWithoutAuthenticationError()
            self.middlewares.insert(0, get_authorization_middleware(self._authorization_strategy))

        if self._authentication_strategy:
            self.middlewares.insert(0, get_authentication_middleware(self._authentication_strategy))

        if self._use_sync_logging:
            self._configure_sync_logging()

        if self._default_headers:
            self.middlewares.insert(0, get_default_headers_middleware(self._default_headers))

        self._normalize_middlewares()

        if self.middlewares:
            self._apply_middlewares_in_routes()

    def apply_routes(self):
        if self._serve_files:
            serve_files(self.router, *self._serve_files)

    async def start(self):
        if self.started:
            return

        if self.on_start:
            await self.on_start.fire()
        self.started = True

        self.apply_routes()
        self.normalize_handlers()
        self.configure_middlewares()

    async def stop(self):
        await self.on_stop.fire()

    async def _handle_lifespan(self, receive, send):
        message = await receive()
        assert message['type'] == 'lifespan.startup'
        await self.start()
        await send({'type': 'lifespan.startup.complete'})

        message = await receive()
        assert message['type'] == 'lifespan.shutdown'
        await self.stop()
        await send({'type': 'lifespan.shutdown.complete'})

    async def after_response(self, request: Request, response: Response):
        """After response callback"""

    async def __call__(self, scope, receive, send):
        if scope['type'] == 'lifespan':
            return await self._handle_lifespan(receive, send)

        # assert scope['type'] == 'http'

        request = Request.incoming(
            scope['method'],
            scope['raw_path'],
            scope['query_string'],
            scope['headers']
        )
        request.scope = scope
        request.content = ASGIContent(receive)

        response = await self.handle(request)
        await send_asgi_response(response, send)

        await self.after_response(request, response)

        request.scope = None
        request.content.dispose()



import asyncio
from urllib.parse import urlencode
from asyncio import TimeoutError
from typing import List, Optional, Union, Type, Any, Callable, Tuple
from .pool import ClientConnectionPools
from .exceptions import *
from blacksheep import (Request,
                        Response,
                        Content,
                        URL,
                        InvalidURL)
from blacksheep.middlewares import get_middlewares_chain
from .connection import ConnectionClosedError
from .cookies import CookieJar, cookies_middleware
from .logs import get_client_logging_middleware


URLType = Union[str, bytes, URL]
Header = Tuple[bytes, bytes]


class RedirectsCache:
    """Used to store permanent redirects urls for later reuse"""

    __slots__ = ('_cache',)

    def __init__(self):
        self._cache = {}

    def store_redirect(self, source, destination):
        self._cache[source] = destination

    def __setitem__(self, key, value):
        self._cache[key] = value

    def __getitem__(self, item):
        try:
            return self._cache[item]
        except KeyError:
            return None

    def __contains__(self, item):
        return item in self._cache


class ClientRequestContext:

    __slots__ = ('path', 'cookies')

    def __init__(self, request, cookies: CookieJar = None):
        self.path = [request.url.value.lower()]
        self.cookies = cookies


def get_default_headers_middleware(headers):
    async def default_headers_middleware(request, handler):
        for header in headers:
            request.headers.add(header)
        return await handler(request)
    return default_headers_middleware


class ClientSession:

    def __init__(self,
                 loop=None,
                 base_url=None,
                 ssl=None,
                 pools=None,
                 default_headers: Optional[List[Header]] = None,
                 follow_redirects: bool = True,
                 connection_timeout: float = 5.0,
                 request_timeout: float = 60.0,
                 maximum_redirects: int = 20,
                 redirects_cache_type: Union[Type[RedirectsCache], Any] = None,
                 cookie_jar: CookieJar = None,
                 middlewares: Optional[List[Callable]] = None):
        if loop is None:
            loop = asyncio.get_event_loop()

        if not pools:
            pools = ClientConnectionPools(loop)

        if redirects_cache_type is None and follow_redirects:
            redirects_cache_type = RedirectsCache

        if middlewares is None:
            middlewares = []

        if cookie_jar is None:
            cookie_jar = CookieJar()

        if cookie_jar:
            middlewares.insert(0, cookies_middleware)

        self.loop = loop
        self._base_url = None
        self.base_url = base_url
        self.ssl = ssl
        self.default_headers = default_headers
        self.pools = pools
        self.connection_timeout = connection_timeout
        self.request_timeout = request_timeout
        self.follow_redirects = follow_redirects
        self.cookie_jar = cookie_jar
        self._permanent_redirects_urls = redirects_cache_type() if follow_redirects else None
        self.non_standard_handling_of_301_302_redirect_method = True
        self.maximum_redirects = maximum_redirects
        self._handler = None
        self._middlewares = middlewares
        if middlewares:
            self._build_middlewares_chain()
        self.delay_before_retry = 0.5

    @property
    def base_url(self):
        return self._base_url

    @base_url.setter
    def base_url(self, value):
        if value and not isinstance(value, URL):
            if isinstance(value, str):
                value = value.encode()
            value = URL(value)
        self._base_url = value

    def add_middlewares(self, middlewares: List[Callable]):
        self._middlewares += middlewares
        self._build_middlewares_chain()

    def set_middlewares(self, middlewares: List[Callable]):
        self._middlewares = middlewares
        self._build_middlewares_chain()

    def _build_middlewares_chain(self):
        async def root_handler(request):
            return await self._send_core(request)

        self._handler = get_middlewares_chain(self._middlewares, root_handler)

    def use_standard_redirect(self):
        """Uses specification compliant handling of 301 and 302 redirects"""
        self.non_standard_handling_of_301_302_redirect_method = False

    def get_url(self, url, params=None) -> bytes:
        value = self._get_url_without_params(url)
        if not params:
            return value
        query = urlencode(params).encode('ascii')
        return value + (b'&' if b'?' in value else b'?') + query

    def _get_url_without_params(self, url) -> bytes:
        if isinstance(url, str):
            url = url.encode()

        if not isinstance(url, URL):
            url = URL(url)

        if url.is_absolute:
            return url.value

        if self.base_url:
            return self.base_url.join(url).value
        return url.value

    def configure(self):
        if self._middlewares and not self._handler:
            self._build_middlewares_chain()

    def use_sync_logging(self):
        self._middlewares.insert(0, get_client_logging_middleware())
        self._build_middlewares_chain()

    async def __aenter__(self):
        self.configure()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()

    async def close(self):
        self.pools.dispose()

    @staticmethod
    def extract_redirect_location(response: Response):
        # if the server returned more than one value, use the first header in order
        location = response.get_first_header(b'Location')
        if not location:
            raise MissingLocationForRedirect(response)
        
        # if the location cannot be parsed as URL, let exception happen: this might be a redirect to a URN!!
        # simply don't follows the redirect, and returns the response to the caller
        try:
            return URL(location)
        except InvalidURL:
            raise UnsupportedRedirect()

    @staticmethod
    def get_redirect_url(request: Request, location: URL):
        if location.is_absolute:
            return location
        # relative redirect URI
        # https://tools.ietf.org/html/rfc7231#section-7.1.2
        return request.url.base_url().join(location)

    def validate_redirect(self, redirect_url: URL, response: Response, context: ClientRequestContext):
        redirect_url_lower = redirect_url.value.lower()
        if redirect_url_lower in context.path:
            context.path.append(redirect_url_lower)

            raise CircularRedirectError(context.path, response)

        context.path.append(redirect_url_lower)

        if len(context.path) > self.maximum_redirects:
            raise MaximumRedirectsExceededError(context.path, response, self.maximum_redirects)

    def update_request_for_redirect(self,
                                    request: Request,
                                    response: Response):
        context = request.context  # type: ClientRequestContext
        status = response.status

        if status == 301 or status == 302:
            if self.non_standard_handling_of_301_302_redirect_method:
                # Change original request method to GET (Browser-like)
                request.method = 'GET'

        if status == 303:
            # 303 See Other
            # Change original request method to GET
            request.method = 'GET'

        location = self.extract_redirect_location(response)
        redirect_url = self.get_redirect_url(request, location)

        if redirect_url.schema.lower() not in {b'http', b'https'}:
            raise UnsupportedRedirect()

        self.validate_redirect(redirect_url, response, context)

        if status == 301 or status == 308:
            self._permanent_redirects_urls[request.url.value] = redirect_url

        request.url = redirect_url

    def merge_default_headers(self, request):
        if not self.default_headers:
            return

        for header in self.default_headers:
            # TODO: support tuple and Header?
            if header[0] not in request.headers:
                request.headers.add(header[0], header[1])

    def check_permanent_redirects(self, request):
        if self.follow_redirects and request.url.value in self._permanent_redirects_urls:
            request.url = self._permanent_redirects_urls[request.url.value]

    async def get_connection(self, url: URL):
        pool = self.pools.get_pool(url.schema, url.host, url.port, self.ssl)

        try:
            return await asyncio.wait_for(pool.get_connection(),
                                          self.connection_timeout)
        except TimeoutError:
            raise ConnectionTimeout(url.base_url(), self.connection_timeout)

    def get_new_context(self, request) -> ClientRequestContext:
        return ClientRequestContext(request, self.cookie_jar)

    def _validate_request_url(self, request: Request):
        if not request.url.is_absolute:
            if self.base_url:
                request.url = URL(self._get_url_without_params(request.url))
            else:
                raise ValueError('request.url must be a complete, absolute URL. Either provide a base_url '
                                 'for the client, or specify a full URL for the request.')

    async def send(self, request: Request):
        self._validate_request_url(request)

        if not hasattr(request, 'context'):
            request.context = self.get_new_context(request)
            self.merge_default_headers(request)

        if self._handler:
            # using middlewares
            return await self._handler(request)

        # without middlewares
        return await self._send_core(request)

    async def _send_core(self, request: Request):
        self.check_permanent_redirects(request)

        response = await self._send_using_connection(request)

        if self.follow_redirects and response.is_redirect():
            try:
                self.update_request_for_redirect(request, response)
            except UnsupportedRedirect:
                # redirect not to HTTP / HTTPS: for example, it can be a redirect to a URN
                return response
            return await self.send(request)

        return response

    async def _send_using_connection(self, request):
        connection = await self.get_connection(request.url)

        try:
            return await asyncio.wait_for(connection.send(request),
                                          self.request_timeout)
        except ConnectionClosedError as connection_closed_error:

            if connection_closed_error.can_retry:
                await asyncio.sleep(self.delay_before_retry, loop=self.loop)
                return await self._send_using_connection(request)
            raise
        except TimeoutError:
            raise RequestTimeout(request.url, self.request_timeout)

    async def get(self,
                  url: URLType,
                  headers: Optional[List[Header]] = None,
                  params=None):
        return await self.send(Request('GET',
                                       self.get_url(url, params),
                                       headers))

    async def post(self,
                   url: URLType,
                   content: Content = None,
                   headers: Optional[List[Header]] = None,
                   params=None):
        return await self.send(Request('POST',
                                       self.get_url(url, params),
                                       headers).with_content(content))

    async def put(self,
                  url: URLType,
                  content: Content = None,
                  headers: Optional[List[Header]] = None,
                  params=None):
        return await self.send(Request('PUT',
                                       self.get_url(url, params),
                                       headers).with_content(content))

    async def delete(self,
                     url: URLType,
                     content: Content = None,
                     headers: Optional[List[Header]] = None,
                     params=None):
        return await self.send(Request('DELETE',
                                       self.get_url(url, params),
                                       headers).with_content(content))

    async def trace(self,
                    url: URLType,
                    headers: Optional[List[Header]] = None,
                    params=None):
        return await self.send(Request('TRACE',
                                       self.get_url(url, params),
                                       headers))

    async def head(self,
                   url: URLType,
                   headers: Optional[List[Header]] = None,
                   params=None):
        return await self.send(Request('HEAD',
                                       self.get_url(url, params),
                                       headers))

    async def patch(self,
                    url: URLType,
                    content: Content = None,
                    headers: Optional[List[Header]] = None,
                    params=None):
        return await self.send(Request('PATCH',
                                       self.get_url(url, params),
                                       headers).with_content(content))

    async def options(self,
                      url: URLType,
                      content: Content = None,
                      headers: Optional[List[Header]] = None,
                      params=None):
        return await self.send(Request('OPTIONS',
                                       self.get_url(url, params),
                                       headers).with_content(content))




import logging
from asyncio import Queue, QueueEmpty, QueueFull
from ssl import SSLContext
from .connection import ClientConnection, SECURE_SSLCONTEXT, INSECURE_SSLCONTEXT
from blacksheep.exceptions import InvalidArgument


logger = logging.getLogger('blacksheep.client')


class ClientConnectionPool:

    def __init__(self, loop, scheme, host, port, ssl=None, max_size=0):
        self.loop = loop
        self.scheme = scheme
        self.host = host
        self.port = port
        self.ssl = self._ssl_option(ssl)
        self.max_size = max_size
        self._idle_connections = Queue(maxsize=max_size)
        self.disposed = False

    def _ssl_option(self, ssl):
        if self.scheme == b'https':
            if ssl is None:
                return SECURE_SSLCONTEXT
            if ssl is False:
                return INSECURE_SSLCONTEXT
            if isinstance(ssl, SSLContext):
                return ssl
            raise InvalidArgument('Invalid ssl argument, expected one of: '
                                  '{None, False, True, instance of ssl.SSLContext}')
        if ssl:
            raise InvalidArgument('SSL argument specified for non-https scheme.')
        return None

    def _get_connection(self):
        # if there are no connections, let QueueEmpty exception happen
        # if all connections are closed, remove all of them and let QueueEmpty exception happen
        while True:
            connection = self._idle_connections.get_nowait()  # type: ClientConnection

            if connection.open:
                logger.debug(f'Reusing connection {id(connection)} to: {self.host}:{self.port}')
                return connection

    def try_return_connection(self, connection):
        if self.disposed:
            return

        try:
            self._idle_connections.put_nowait(connection)
        except QueueFull:
            pass

    async def get_connection(self):
        try:
            return self._get_connection()
        except QueueEmpty:
            return await self.create_connection()

    async def create_connection(self):
        logger.debug(f'Creating connection to: {self.host}:{self.port}')
        transport, connection = await self.loop.create_connection(
            lambda: ClientConnection(self.loop, self),
            self.host,
            self.port,
            ssl=self.ssl)
        await connection.ready.wait()
        # NB: a newly created connection is going to be used by a request-response cycle;
        # so we don't put it inside the pool (since it's not immediately reusable for other requests)
        return connection

    def dispose(self):
        self.disposed = True
        while True:
            try:
                connection = self._idle_connections.get_nowait()
            except QueueEmpty:
                break
            else:
                logger.debug(f'Closing connection {id(connection)} to: {self.host}:{self.port}')
                connection.close()


class ClientConnectionPools:

    def __init__(self, loop):
        self.loop = loop
        self._pools = {}

    def get_pool(self, scheme, host, port, ssl):
        assert scheme in (b'http', b'https'), 'URL schema must be http or https'
        if port is None or port == 0:
            port = 80 if scheme == b'http' else 443
        
        key = (scheme, host, port)
        try:
            return self._pools[key]
        except KeyError:
            new_pool = ClientConnectionPool(self.loop, scheme, host, port, ssl)
            self._pools[key] = new_pool
            return new_pool

    def dispose(self):
        for key, pool in self._pools.items():
            pool.dispose()



cimport cython
from .exceptions cimport MessageAborted

import re
import uuid
import json
from inspect import isasyncgenfunction
from collections.abc import MutableSequence
from typing import Union, List, Optional, Callable, Any
from urllib.parse import parse_qsl
from urllib.parse import quote_plus


_charset_rx = re.compile(b'charset=\\s?([^;]+)')
_boundary_rx = re.compile(b'boundary=(.+)$', re.I)
_content_type_form_data_line_rx = re.compile(b'content-type:\\s?([^;]+)(?:(charset=\\s?([^;]+)))?', re.I)
_content_disposition_header_type_rx = re.compile(b'^content-disposition:\\s([^;]+)', re.I)
_content_disposition_header_name_rx = re.compile(b'\\sname="([^\\"]+)"', re.I)
_content_disposition_header_filename_rx = re.compile(b'\\sfilename="([^\\"]+)"', re.I)


cdef class Content:

    def __init__(self,
                 bytes content_type,
                 bytes data):
        self.type = content_type
        self.body = data
        self.length = len(data)

    async def read(self):
        return self.body


cdef class StreamedContent(Content):

    def __init__(self,
                 bytes content_type,
                 object data_provider):
        self.type = content_type
        self.body = None
        self.length = -1
        self.generator = data_provider
        if not isasyncgenfunction(data_provider):
            raise ValueError('Data provider must be an async generator')

    async def get_parts(self):
        async for chunk in self.generator():
            yield chunk


cdef class ASGIContent(Content):

    def __init__(self, object receive):
        self.type = None
        self.body = None
        self.length = -1
        self.receive = receive

    cpdef void dispose(self):
        self.receive = None
        self.body = None

    async def stream(self):
        while True:
            message = await self.receive()

            if message.get('type') == 'http.disconnect':
                raise MessageAborted()

            yield message.get('body', b'')

            if not message.get('more_body'):
                break

        yield b''

    async def read(self):
        if self.body is not None:
            return self.body
        value = bytearray()

        while True:
            message = await self.receive()

            if message.get('type') == 'http.disconnect':
                raise MessageAborted()

            value.extend(message.get('body', b''))

            if not message.get('more_body'):
                break

        self.body = bytes(value)
        self.length = len(self.body)
        return self.body


cdef class TextContent(Content):

    def __init__(self, str text):
        super().__init__(b'text/plain; charset=utf-8', text.encode('utf8'))


cdef class HtmlContent(Content):

    def __init__(self, str html):
        super().__init__(b'text/html; charset=utf-8', html.encode('utf8'))


cdef class JsonContent(Content):

    def __init__(self, object data, dumps=json.dumps):
        super().__init__(b'application/json', dumps(data).encode('utf8'))


cdef dict parse_www_form_urlencoded(str content):
    # application/x-www-form-urlencoded
    cdef str key, value
    cdef dict data = {}
    for key, value in parse_qsl(content):
        if key in data:
            data[key].append(value)
        else:
            data[key] = [value]
    return data


cpdef dict parse_www_form(str content):
    return parse_www_form_urlencoded(content)


cpdef bytes extract_multipart_form_data_boundary(bytes content_type):
    m = _boundary_rx.search(content_type)
    if not m:
        return None
    return m.group(1)


cpdef tuple parse_content_disposition_header(bytes header):
    # content-disposition: form-data; name="file1"; filename="a.txt"
    type_m = _content_disposition_header_type_rx.search(header)

    if not type_m:
        raise ValueError(f'Failed to parse content-disposition type: {header}')

    name_m = _content_disposition_header_name_rx.search(header)
    if not name_m:
        raise ValueError(f'Given content-disposition header does not contain required field name: {header}')

    file_name_m = _content_disposition_header_filename_rx.search(header)
    return type_m.group(1), name_m.group(1), file_name_m.group(1) if file_name_m else None


cpdef bytes remove_last_crlf(bytes data):
    if data[-2:] == b'\r\n':
        return data[:-2]
    if data[-1:] == b'\n':
        return data[:-1]
    return data


cpdef bytes remove_first_crlf(bytes data):
    if data[:2] == b'\r\n':
        return data[2:]
    if data[:1] == b'\n':
        return data[1:]
    return data


cpdef bytes remove_extreme_crlf(bytes data):
    return remove_last_crlf(remove_first_crlf(data))


cpdef list split_multipart_form_data_parts(bytes data):
    result = []
    cdef int j = 0
    cdef int k = 0
    cdef char c

    for i, c in enumerate(data):
        if c == 10:
            k += 1
            if k == 3:
                result.append(remove_extreme_crlf(data[j:]))
                return result

            result.append(data[j:i])
            j = i+1


cpdef list parse_multipart_form_data(bytes content, bytes boundary):
    cdef bytes part
    cdef list result = []
    cdef bytes default_charset = b'utf8'
    cdef:
        bytes charset
        bytes content_type
        bytes disposition_type
        bytes name
        bytes file_name

    for part in content.split(boundary):
        if part == b'' or part == b'--' or part == b'--\r' or part == b'--\r\n':
            continue

        parts = split_multipart_form_data_parts(part.lstrip(b'\r\n'))

        charset = None
        content_type = None
        disposition_type, name, file_name = parse_content_disposition_header(parts[0])

        if name == b'_charset_':
            # https://tools.ietf.org/html/rfc7578#section-4.6
            default_charset = parts[2]
            continue

        # NB: if a content-type is defined, the second line is populated, otherwise is an empty bytes b''
        if parts[1]:
            content_type_m = _content_type_form_data_line_rx.search(parts[1])
            if content_type_m:
                content_type = content_type_m.group(1)
                charset = content_type_m.group(2)
                if not charset and content_type == b'text/plain':
                    charset = default_charset

        result.append(FormPart(name, parts[2], content_type, file_name, charset))
    return result


cpdef void write_multipart_part(FormPart part, bytearray destination):
    # https://tools.ietf.org/html/rfc7578#page-4
    destination.extend(b'Content-Disposition: form-data; name="')
    destination.extend(part.name)
    destination.extend(b'"')
    if part.file_name:
        destination.extend(b'; filename="')
        destination.extend(part.file_name)
        destination.extend(b'"\r\n')
    if part.content_type:
        destination.extend(b'Content-Type: ')
        destination.extend(part.content_type)
    destination.extend(b'\r\n\r\n')
    destination.extend(part.data)
    destination.extend(b'\r\n')


cpdef bytes write_www_form_urlencoded(data: Union[dict, list]):
    # application/x-www-form-urlencoded
    if isinstance(data, list):
        values = data
    else:
        values = data.items()
    
    cdef list contents = []
    
    for key, value in values:
        if isinstance(value, MutableSequence):
            for item in value:
                contents.append(quote_plus(key) + '=' + quote_plus(str(item)))
        else:
            contents.append(quote_plus(key) + '=' + quote_plus(str(value)))
    return ('&'.join(contents)).encode('utf8')


cdef class FormContent(Content):

    def __init__(self, data: dict):
        super().__init__(b'application/x-www-form-urlencoded', write_www_form_urlencoded(data))


cdef class FormPart:

    def __init__(self,
                 bytes name,
                 bytes data,
                 bytes content_type: Optional[bytes]=None,
                 bytes file_name: Optional[bytes]=None,
                 bytes charset: Optional[bytes] = None):
        self.name = name
        self.data = data
        self.file_name = file_name
        self.content_type = content_type
        self.charset = charset

    def __repr__(self):
        return f'<FormPart {self.name} - at {id(self)}>'


cdef class MultiPartFormData(Content):

    def __init__(self, list parts):
        self.parts = parts
        self.boundary = b'------' + str(uuid.uuid4()).replace('-', '').encode()
        super().__init__(b'multipart/form-data; boundary=' + self.boundary, write_multipart_form_data(self))


cpdef bytes write_multipart_form_data(MultiPartFormData data):
    cdef bytearray contents = bytearray()
    cdef FormPart part
    for part in data.parts:
        contents.extend(b'--')
        contents.extend(data.boundary)
        contents.extend(b'\r\n')
        write_multipart_part(part, contents)
    contents.extend(b'--')
    contents.extend(data.boundary)
    contents.extend(b'--\r\n')
    return bytes(contents)







from functools import lru_cache
from blacksheep.server import Application
from blacksheep import Response, Content
from jinja2 import Environment, Template, PackageLoader, select_autoescape


@lru_cache(1200)
def template_name(name: str):
    if not name.endswith('.html'):
        return name + '.html'
    return name


def get_response(html: str):
    return Response(200, [
        (b'Cache-Control', b'no-cache')
    ]).with_content(Content(b'text/html; charset=utf-8', html.encode('utf8')))


def render_template(template: Template, *args, **kwargs):
    return template.render(*args, **kwargs)


async def render_template_async(template: Template, *args, **kwargs):
    return await template.render_async(*args, **kwargs)


def use_templates(app: Application, loader: PackageLoader, enable_async: bool = False):
    env = app.services['jinja_environment']
    if not env:
        env = Environment(
            loader=loader,
            autoescape=select_autoescape(['html', 'xml']),
            auto_reload=app.debug,
            enable_async=enable_async
        )

        app.services['jinja_environment'] = env
        app.services['jinja'] = env
        env.globals['app'] = app

    if enable_async:
        async def async_view(name: str, *args, **kwargs):
            return get_response(await render_template_async(env.get_template(template_name(name)), *args, **kwargs))

        return async_view

    def sync_view(name: str, *args, **kwargs):
        return get_response(render_template(env.get_template(template_name(name)), *args, **kwargs))

    return sync_view


def view(jinja_environment, name: str, *args, **kwargs):
    """Returns a Response object with HTML obtained from synchronous rendering.

    Use this when `enable_async` is set to False when calling `use_templates`."""
    return get_response(render_template(jinja_environment.get_template(template_name(name)),
                                        *args, **kwargs))


async def view_async(jinja_environment, name: str, *args, **kwargs):
    """Returns a Response object with HTML obtained from synchronous rendering."""
    return get_response(await render_template_async(jinja_environment
                                                    .get_template(template_name(name)), *args, **kwargs))



"""
This module implements a feature inspired by "Model Binding" in ASP.NET web framework.
It provides a strategy to have request parameters read an injected into request handlers calls.
This feature is also useful to generate OpenAPI Documentation (Swagger) automatically (not implemented, yet).

See:
    https://docs.microsoft.com/en-us/aspnet/core/mvc/models/model-binding?view=aspnetcore-2.2
"""
from abc import ABC, abstractmethod
from collections.abc import Iterable as IterableAbc
from typing import Type, TypeVar, Optional, Callable, Sequence, Union, List
from urllib.parse import unquote
from blacksheep import Request
from blacksheep.exceptions import BadRequest
from rodi import Services, GetServiceContext


T = TypeVar('T')
TypeOrName = Union[Type, str]


def _inspect_is_list_typing(expected_type):
    return hasattr(expected_type, '__origin__') and expected_type.__origin__ is list


def _generalize_init_type_error_message(ex: TypeError) -> str:
    return str(ex)\
        .replace('__init__() ', '')\
        .replace('keyword argument', 'parameter') \
        .replace('keyword arguments', 'parameters') \
        .replace('positional arguments', 'parameters')\
        .replace('positional argument', 'parameter')


class Binder(ABC):

    def __init__(self,
                 expected_type: T,
                 required: bool = True,
                 converter: Optional[Callable] = None):
        self.expected_type = expected_type
        self.required = required
        self.converter = converter
        self.name = None

    @abstractmethod
    async def get_value(self, request: Request) -> T:
        pass

    def __repr__(self):
        return f'<{self.__class__.__name__} {self.expected_type} at {id(self)}>'


class MissingBodyError(BadRequest):

    def __init__(self):
        super().__init__('Missing body payload')


class MissingParameterError(BadRequest):

    def __init__(self, name: str, source: str):
        super().__init__(f'Missing parameter `{name}` from {source}')


class InvalidRequestBody(BadRequest):

    def __init__(self, description: Optional[str] = 'Invalid body payload'):
        super().__init__(description)


class MissingConverterError(Exception):

    def __init__(self, expected_type, binder_type):
        super().__init__(f'A default converter for type `{str(expected_type)}` is not configured. '
                         f'Please define a converter method for this binder ({binder_type.__name__}).')


class FromBody(Binder):

    _excluded_methods = {'GET', 'HEAD', 'TRACE'}


class FromJson(FromBody):

    def __init__(self,
                 expected_type: T,
                 required: bool = False,
                 converter: Optional[Callable] = None
                 ):
        if not converter:
            def default_converter(data):
                return expected_type(**data)
            converter = default_converter

        super().__init__(expected_type, required, converter)

    def parse_value(self, data: dict) -> T:
        try:
            return self.converter(data)
        except TypeError as te:
            raise InvalidRequestBody(_generalize_init_type_error_message(te))
        except ValueError as ve:
            raise InvalidRequestBody(str(ve))

    async def get_value(self, request: Request) -> T:
        if request.declares_json() and request.method not in self._excluded_methods:
            data = await request.json()

            if not data:
                raise MissingBodyError()

            return self.parse_value(data)

        if self.required:
            if not request.has_body():
                raise MissingBodyError()

            raise InvalidRequestBody('Expected JSON payload')

        return None


def _default_bool_converter(value: str):
    if value in {'1', 'true'}:
        return True

    if value in {'0', 'false'}:
        return False

    # bad request: expected a bool value, but
    # got something different that is not handled
    raise BadRequest()


def _default_bool_list_converter(values: Sequence[str]):
    return _default_bool_converter(values[0].lower()) if values else None


def _default_str_binder(values: Sequence[str]):
    if not values:
        return None
    return unquote(values[0])


def _default_simple_types_binder(expected_type: Type, values: Sequence[str]):
    return expected_type(values[0]) if values else None


class SyncBinder(Binder):
    """Base binder class for values that can be read synchronously from requests with complete headers.
    Like route, query string and header parameters.
    """

    _simple_types = {int, float, bool}

    def __init__(self,
                 expected_type: T = List[str],
                 name: str = None,
                 required: bool = False,
                 converter: Optional[Callable] = None):
        super().__init__(expected_type, required, converter or self._get_default_converter(expected_type))
        self.name = name

    def __repr__(self):
        return f'<{self.__class__.__name__} "{self.name}" at {id(self)}>'

    def _get_default_converter_single(self, expected_type):
        if expected_type is str:
            return lambda value: unquote(value) if value else None

        if expected_type is bool:
            return _default_bool_converter

        if expected_type is bytes:
            return lambda value: value if value else None

        if expected_type in self._simple_types:
            return lambda value: expected_type(value) if value else None

        raise MissingConverterError(expected_type, self.__class__)

    def _get_default_converter_for_iterable(self, expected_type):
        generic_type = self._get_type_for_generic_iterable(expected_type)
        item_type = self._generic_iterable_annotation_item_type(expected_type)
        item_converter = self._get_default_converter_single(item_type)
        return lambda values: generic_type(item_converter(value) for value in values)

    def _get_default_converter(self, expected_type):
        if expected_type is str:
            return lambda value: unquote(value[0]) if value else None

        if expected_type is bool:
            return _default_bool_list_converter

        if expected_type is bytes:
            return lambda value: value[0] if value else None

        if expected_type in self._simple_types:
            return lambda value: expected_type(value[0]) if value else None

        if self._is_generic_iterable_annotation(expected_type) or expected_type in {list, set, tuple}:
            return self._get_default_converter_for_iterable(expected_type)

        raise MissingConverterError(expected_type, self.__class__)

    def _get_type_for_generic_iterable(self, expected_type):
        if expected_type in {list, tuple, set}:
            return expected_type

        origin = expected_type.__origin__
        if origin in {list, tuple, set}:
            return origin
        # here we cannot make something perfect: if the user of the library wants something better,
        # a converter should be specified when configuring binders; here the code defaults to list
        # for all abstract types (typing.Sequence, Set, etc.) even though not perfect
        return list

    def _is_generic_iterable_annotation(self, param_type):
        return hasattr(param_type, '__origin__') and (param_type.__origin__ in {list, tuple, set}
                                                      or issubclass(param_type.__origin__, IterableAbc))

    def _generic_iterable_annotation_item_type(self, param_type):
        try:
            item_type = param_type.__args__[0]
        except (IndexError, AttributeError):
            return str

        if isinstance(item_type, TypeVar):
            return str
        return item_type

    @abstractmethod
    def get_raw_value(self, request: Request) -> Sequence[str]:
        pass

    @property
    @abstractmethod
    def source_name(self) -> str:
        pass

    _empty_iterables = [list(), set(), tuple()]

    def _empty_iterable(self, value):
        return value in self._empty_iterables

    async def get_value(self, request: Request) -> T:
        raw_value = self.get_raw_value(request)
        try:
            value = self.converter(raw_value)
        except (ValueError, BadRequest):
            raise BadRequest(f'Invalid value for parameter `{self.name}`; expected {self.expected_type}')

        if value is None and self.required:
            raise MissingParameterError(self.name, self.source_name)

        if not self.required and self._empty_iterable(value):
            return None

        return value


class FromHeader(SyncBinder):

    @property
    def source_name(self) -> str:
        return 'header'

    def get_raw_value(self, request: Request) -> Sequence[str]:
        return [header.decode('utf8') for header in request.get_headers(self.name.encode())]


class FromQuery(SyncBinder):

    @property
    def source_name(self) -> str:
        return 'query'

    def get_raw_value(self, request: Request) -> Sequence[str]:
        return [value for value in request.query.get(self.name, [])]


class FromRoute(SyncBinder):

    def __init__(self,
                 expected_type: T = str,
                 name: str = None,
                 required: bool = False,
                 converter: Optional[Callable] = None):
        super().__init__(expected_type, name, required, converter)

    def get_raw_value(self, request: Request) -> Sequence[str]:
        return [request.route_values.get(self.name, '')]

    @property
    def source_name(self) -> str:
        return 'route'


class FromServices(Binder):

    def __init__(self, service: TypeOrName, services: Optional[Services] = None):
        super().__init__(service, False, None)
        self.services = services

    async def get_value(self, request: Request) -> T:
        try:
            context = request.services_context
        except AttributeError:
            # no support for scoped services (across parameters and middlewares)
            context = None

        return self.services.get(self.expected_type, context)


class RequestBinder(Binder):

    def __init__(self):
        super().__init__(Request)

    async def get_value(self, request: Request) -> T:
        return request


class ExactBinder(Binder):

    def __init__(self, exact_object):
        super().__init__(object)
        self.exact_object = exact_object

    async def get_value(self, request: Request):
        return self.exact_object



from blacksheep import URL


class InvalidResponseException(Exception):

    def __init__(self, message, response):
        super().__init__(message)
        self.response = response


class MissingLocationForRedirect(InvalidResponseException):

    def __init__(self, response):
        super().__init__(f'The server returned a redirect status ({response.status}) '
                         f'but didn`t send a "Location" header', response)


class RequestException(Exception):

    def __init__(self, message, allow_retry):
        super().__init__(message)
        self.can_retry = allow_retry


class ConnectionTimeout(TimeoutError):
    def __init__(self, url: URL, timeout: float):
        super().__init__(f'Connection attempt timed out, to {url.value.decode()}. Current timeout setting: {timeout}.')


class RequestTimeout(TimeoutError):
    def __init__(self, url: URL, timeout: float):
        super().__init__(f'Request timed out, to: {url.value.decode()}. Current timeout setting: {timeout}.')


class CircularRedirectError(InvalidResponseException):

    def __init__(self, path, response):
        path_string = ' --> '.join(x.decode('utf8') for x in path)
        super().__init__(f'Circular redirects detected. Requests path was: ({path_string}).', response)


class MaximumRedirectsExceededError(InvalidResponseException):

    def __init__(self, path, response, maximum_redirects):
        path_string = ', '.join(x.decode('utf8') for x in path)
        super().__init__(f'Maximum Redirects Exceeded ({maximum_redirects}). Requests path was: ({path_string}).',
                         response)


class UnsupportedRedirect(Exception):
    """Exception risen when the client cannot handle a redirect;
    for example if the redirect is to a URN (not a URL). In such case,
    we don't follow the redirect and return the response with location: the caller
    can handle it."""



from rodi import GetServiceContext


async def dependency_injection_middleware(request, handler):
    with GetServiceContext() as context:
        request.services_context = context
        return await handler(request)



# cython: language_level=3, embedsignature=True
# Copyright (C) 2018-present Roberto Prevato
#
# This module is part of BlackSheep and is released under
# the MIT License https://opensource.org/licenses/MIT


cdef class Cookie:
    cdef object _expiration
    cdef public bytes name
    cdef public bytes value
    cdef public bytes expires
    cdef public bytes domain
    cdef public bytes path
    cdef public bint http_only
    cdef public bint secure
    cdef public bytes max_age
    cdef public bytes same_site
    cpdef Cookie clone(self)
    cpdef void set_max_age(self, int max_age)


cpdef Cookie parse_cookie(bytes value)


cpdef bytes datetime_to_cookie_format(object value)


cpdef object datetime_from_cookie_format(bytes value)


cdef bytes write_cookie_for_response(Cookie cookie)



from setuptools import setup, Extension


def readme():
    with open('README.md') as f:
        return f.read()


COMPILE_ARGS = ['-O3']


setup(name='blacksheep',
      version='0.1.8',
      description='Fast web framework and HTTP client for Python asyncio',
      long_description=readme(),
      long_description_content_type='text/markdown',
      classifiers=[
          'Development Status :: 4 - Beta',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python :: 3.7',
          'Operating System :: OS Independent',
          'Framework :: AsyncIO'
      ],
      url='https://github.com/RobertoPrevato/BlackSheep',
      author='Roberto Prevato',
      author_email='roberto.prevato@gmail.com',
      keywords='BlackSheep web framework',
      platforms=['*nix'],
      license='MIT',
      packages=['blacksheep',
                'blacksheep.server',
                'blacksheep.server.files',
                'blacksheep.server.res',
                'blacksheep.client'],
      ext_modules=[
          Extension('blacksheep.url',
                    ['blacksheep/url.c'],
                    extra_compile_args=COMPILE_ARGS),

          Extension('blacksheep.exceptions',
                    ['blacksheep/exceptions.c'],
                    extra_compile_args=COMPILE_ARGS),

          Extension('blacksheep.headers',
                    ['blacksheep/headers.c'],
                    extra_compile_args=COMPILE_ARGS),

          Extension('blacksheep.cookies',
                    ['blacksheep/cookies.c'],
                    extra_compile_args=COMPILE_ARGS),

          Extension('blacksheep.contents',
                    ['blacksheep/contents.c'],
                    extra_compile_args=COMPILE_ARGS),

          Extension('blacksheep.messages',
                    ['blacksheep/messages.c'],
                    extra_compile_args=COMPILE_ARGS),

          Extension('blacksheep.scribe',
                    ['blacksheep/scribe.c'],
                    extra_compile_args=COMPILE_ARGS),

          Extension('blacksheep.baseapp',
                    ['blacksheep/baseapp.c'],
                    extra_compile_args=COMPILE_ARGS)
      ],
      install_requires=[
          'httptools',
          'certifi',
          'cchardet',
          'guardpost',
          'rodi',
          'aiofiles'
      ],
      include_package_data=True,
      zip_safe=False)



import json
import ntpath
import aiofiles
from io import BytesIO
from enum import Enum
from functools import lru_cache
from typing import Any, Callable, Union
from blacksheep import Response, TextContent, HtmlContent, JsonContent, Content, StreamedContent


class ContentDispositionType(Enum):
    INLINE = 'inline'
    ATTACHMENT = 'attachment'


def _ensure_bytes(value: Union[str, bytes]):
    if isinstance(value, str):
        return value.encode()
    if isinstance(value, bytes):
        return value
    raise ValueError('Input value must be bytes or str')


def status_code(status: int = 200, message: Union[None, str, dict] = None):
    """Returns a plain response with given status, with optional message; sent as plain text or JSON."""
    if not message:
        return Response(status)
    if isinstance(message, str):
        content = TextContent(message)
    else:
        content = JsonContent(message)
    return Response(status, content=content)


def ok(message: Union[None, str, dict] = None):
    """Returns an HTTP 200 OK response, with optional message; sent as plain text or JSON."""
    return status_code(200, message)


def created(location: Union[bytes, str], value: Any = None):
    """Returns an HTTP 201 Created response, to the given location and with optional JSON content."""
    return Response(201,
                    [(b'Location', _ensure_bytes(location))],
                    JsonContent(value) if value else None)


def accepted():
    """Returns an HTTP 202 Accepted response."""
    return Response(202)


def no_content():
    """Returns an HTTP 204 No Content response."""
    return Response(204)


def not_modified():
    """Returns an HTTP 304 Not Modified response."""
    return Response(304)


def unauthorized(message: Union[None, str, dict] = None):
    """Returns an HTTP 401 Unauthorized response, with optional message; sent as plain text or JSON."""
    return status_code(401, message)


def forbidden(message: Union[None, str, dict] = None):
    """Returns an HTTP 403 Forbidden response, with optional message; sent as plain text or JSON."""
    return status_code(403, message)


def bad_request(message: Union[None, str, dict] = None):
    """Returns an HTTP 400 Bad Request response, with optional message; sent as plain text or JSON."""
    return status_code(400, message)


def not_found():
    """Returns an HTTP 404 Not Found response"""
    return Response(404)


def moved_permanently(location: Union[bytes, str]):
    """Returns an HTTP 301 Moved Permanently response, to the given location"""
    return Response(301, [(b'Location', _ensure_bytes(location))])


def redirect(location: Union[bytes, str]):
    """Returns an HTTP 302 Found response (commonly called redirect), to the given location"""
    return Response(302, [(b'Location', _ensure_bytes(location))])


def see_other(location: Union[bytes, str]):
    """Returns an HTTP 303 See Other response, to the given location."""
    return Response(303, [(b'Location', _ensure_bytes(location))])


def temporary_redirect(location: Union[bytes, str]):
    """Returns an HTTP 307 Temporary Redirect response, to the given location."""
    return Response(307, [(b'Location', _ensure_bytes(location))])


def permanent_redirect(location: Union[bytes, str]):
    """Returns an HTTP 308 Permanent Redirect response, to the given location."""
    return Response(308, [(b'Location', _ensure_bytes(location))])


def text(value: str, status: int = 200):
    """Returns a response with text/plain content, and given status (default HTTP 200 OK)."""
    return Response(status, None, Content(b'text/plain; charset=utf-8', value.encode('utf8')))


def html(value: str, status: int = 200):
    """Returns a response with text/html content, and given status (default HTTP 200 OK)."""
    return Response(status, None, Content(b'text/html; charset=utf-8', value.encode('utf8')))


def json(data: Any, status: int = 200, dumps=json.dumps):
    """Returns a response with application/json content, and given status (default HTTP 200 OK)."""
    return Response(status, None, Content(b'application/json', dumps(data).encode('utf8')))


FileInput = Union[Callable, str, bytes, bytearray, BytesIO]


@lru_cache(2000)
def _get_file_provider(file_path: str):
    async def data_provider():
        async with aiofiles.open(file_path, mode='rb') as f:
            while True:
                chunk = await f.read(1024 * 64)

                if not chunk:
                    break

                yield chunk
            yield b''
    return data_provider


def _file(value: FileInput,
          content_type: Union[str, bytes],
          content_disposition_type: ContentDispositionType,
          file_name: str = None):
    if file_name:
        exact_file_name = ntpath.basename(file_name)
        if not exact_file_name:
            raise ValueError('Invalid file name: it should be an exact file name without path, for example: "foo.txt"')

        content_disposition_value = f'{content_disposition_type.value}; filename="{exact_file_name}"'
    else:
        content_disposition_value = content_disposition_type.value

    content_type = _ensure_bytes(content_type)

    if isinstance(value, str):
        # value is treated as a path
        content = StreamedContent(content_type, _get_file_provider(value))
    elif isinstance(value, BytesIO):
        async def data_provider():
            while True:
                chunk = value.read(1024*64)

                if not chunk:
                    break

                yield chunk
            yield b''

        content = StreamedContent(content_type, data_provider)
    elif callable(value):
        # value is treated as an async generator
        async def data_provider():
            async for chunk in value():
                yield chunk
            yield b''

        content = StreamedContent(content_type, data_provider)
    elif isinstance(value, bytes):
        content = Content(content_type, value)
    elif isinstance(value, bytearray):
        content = Content(content_type, bytes(value))
    else:
        raise ValueError('Invalid value, expected one of: Callable, str, bytes, bytearray, io.BytesIO')

    return Response(200, [(b'Content-Disposition', content_disposition_value.encode())], content)


def inline_file(value: FileInput,
                content_type: Union[str, bytes],
                file_name: str = None):
    """Returns a binary file response with given content type and optional file name, for inline use
    (default HTTP 200 OK). This method supports both call with bytes, or a generator yielding chunks.

    Remarks: this method does not handle cache, ETag and HTTP 304 Not Modified responses;
    when handling files it is recommended to handle cache, ETag and Not Modified, according to use case."""
    return _file(value, content_type, ContentDispositionType.INLINE, file_name)


def file(value: FileInput,
         content_type: Union[str, bytes],
         file_name: str = None,
         content_disposition: ContentDispositionType = ContentDispositionType.ATTACHMENT):
    """Returns a binary file response with given content type and optional file name, for download (attachment)
    (default HTTP 200 OK). This method supports both call with bytes, or a generator yielding chunks.

    Remarks: this method does not handle cache, ETag and HTTP 304 Not Modified responses;
    when handling files it is recommended to handle cache, ETag and Not Modified, according to use case."""
    return _file(value, content_type, content_disposition, file_name)

